@@ -28,10 +28,9 @@ CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "options.h"
 
 gib_list *filelist = NULL;
+int       filelist_len = 0;
 gib_list *current_file = NULL;
 extern int errno;
-int file_num = 0;
-
 
 static gib_list *rm_filelist = NULL;
 
@@ -517,12 +516,24 @@ feh_read_filelist(char *filename)
    FILE *fp;
    gib_list *list = NULL;
    char s[1024], s1[1024];
+   Imlib_Image im1;
 
    D_ENTER(4);
 
    if (!filename)
       D_RETURN(4, NULL);
 
+   /* try and load the given filelist as an image, cowardly refuse to
+    * overwrite an image with a filelist. (requested by user who did feh -df *
+    * when he meant feh -dF *, as it overwrote the first image with the
+    * filelist).
+    */
+   if (feh_load_image_char(&im1, filename)) {
+      weprintf("The file you specified as a filelist to read - %s - appears to be an image. Ignoring it (this is a common mistake).\n", filename);
+      opt.filelistfile = NULL;
+      D_RETURN(4, NULL);
+   }
+   
    errno = 0;
    if ((fp = fopen(filename, "r")) == NULL)
    {
